"""Agent asset."""
import dataclasses
from typing import Optional

from ostorlab.assets import asset


@dataclasses.dataclass
@asset.selector('v3.asset.agent')
class Agent(asset.Asset):
    """Agent asset."""

    def __init__(self, key: str, version: Optional[str] = None, git_location: Optional[str] = None,
                 docker_location: Optional[str] = None,
                 yaml_file_location: Optional[str] = None):
        self.key = key
        self.version = version
        self.git_location = git_location
        self.docker_location = docker_location
        self.yaml_file_location = yaml_file_location

    def __str__(self) -> str:
        if self.version is not None:
            return f'Agent {self.key}:{self.version}'
        else:
            return f'Agent {self.key}'

    @property
    def proto_field(self) -> str:
        return 'agent'



"""Create Scan via an API Request."""
import json
from typing import Dict, Optional

from ostorlab.apis import request


class CreateAgentScanAPIRequest(request.APIRequest):
    """Persist scan API request."""

    def __init__(self, title: str, asset_id: int, agent_group_id: int) -> None:
        """Initializer"""
        self._title = title
        self.asset_id = asset_id
        self._agent_group_id = agent_group_id


    @property
    def query(self) -> Optional[str]:
        """Sets the query of the API request.

        Returns:
            The query to create the scan.
        """
        return """
            mutation StartAgentScan($scan: AgentScanInputType!){
                createAgentScan(scan: $scan) {
                    scan{
                        id
                    }
                }
            }
        """


    @property
    def data(self) -> Optional[Dict]:
        """Sets the body of the API request.

        Returns:
            The body of the create scan request.
        """
        variables = {
            'scan': {
                'title': self._title,
                'assetId': self.asset_id,
                'agentGroupId': self._agent_group_id
            }
        }
        data = {
            'query': self.query,
            'variables': json.dumps(variables)
        }
        return data



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ostorlab/agent/message/proto/v3/asset/link/link.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n5ostorlab/agent/message/proto/v3/asset/link/link.proto\x12*ostorlab.agent.message.proto.v3.asset.link\"%\n\x06Header\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"%\n\x06\x43ookie\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"7\n\x0cLocalStorage\x12\x0b\n\x03url\x18\x01 \x01(\t\x12\x0b\n\x03key\x18\x02 \x01(\t\x12\r\n\x05value\x18\x03 \x01(\t\"6\n\tFormInput\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\x12\x0c\n\x04type\x18\x03 \x01(\t\"^\n\x04\x46orm\x12\x45\n\x06inputs\x18\x01 \x03(\x0b\x32\x35.ostorlab.agent.message.proto.v3.asset.link.FormInput\x12\x0f\n\x07\x65nctype\x18\x02 \x01(\t\">\n\x0e\x46ormCredential\x12\r\n\x05login\x18\x01 \x01(\t\x12\x10\n\x08password\x18\x02 \x01(\t\x12\x0b\n\x03url\x18\x03 \x01(\t\"\"\n\x10ScriptCredential\x12\x0e\n\x06script\x18\x01 \x01(\t\"\xf7\x04\n\x07Message\x12\x0b\n\x03url\x18\x01 \x01(\t\x12\x0e\n\x06method\x18\x02 \x01(\t\x12\x0e\n\x04\x62ody\x18\x03 \x01(\x0cH\x00\x12@\n\x04\x66orm\x18\x04 \x01(\x0b\x32\x30.ostorlab.agent.message.proto.v3.asset.link.FormH\x00\x12I\n\rextra_headers\x18\x05 \x03(\x0b\x32\x32.ostorlab.agent.message.proto.v3.asset.link.Header\x12\x43\n\x07\x63ookies\x18\x06 \x03(\x0b\x32\x32.ostorlab.agent.message.proto.v3.asset.link.Cookie\x12U\n\x0f\x66orm_credential\x18\x07 \x01(\x0b\x32:.ostorlab.agent.message.proto.v3.asset.link.FormCredentialH\x01\x12Y\n\x11script_credential\x18\x08 \x01(\x0b\x32<.ostorlab.agent.message.proto.v3.asset.link.ScriptCredentialH\x01\x12\x43\n\x06parent\x18\t \x01(\x0b\x32\x33.ostorlab.agent.message.proto.v3.asset.link.Message\x12T\n\x12localstorage_items\x18\n \x03(\x0b\x32\x38.ostorlab.agent.message.proto.v3.asset.link.LocalStorageB\x0c\n\nbody_oneofB\x12\n\x10\x63redential_oneof')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ostorlab.agent.message.proto.v3.asset.link.link_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _HEADER._serialized_start=101
  _HEADER._serialized_end=138
  _COOKIE._serialized_start=140
  _COOKIE._serialized_end=177
  _LOCALSTORAGE._serialized_start=179
  _LOCALSTORAGE._serialized_end=234
  _FORMINPUT._serialized_start=236
  _FORMINPUT._serialized_end=290
  _FORM._serialized_start=292
  _FORM._serialized_end=386
  _FORMCREDENTIAL._serialized_start=388
  _FORMCREDENTIAL._serialized_end=450
  _SCRIPTCREDENTIAL._serialized_start=452
  _SCRIPTCREDENTIAL._serialized_end=486
  _MESSAGE._serialized_start=489
  _MESSAGE._serialized_end=1120
# @@protoc_insertion_point(module_scope)



"""This module is the entry point for ostorlab CLI."""
import logging
from typing import Optional

import click

from ostorlab import configuration_manager

logger = logging.getLogger('CLI')


@click.group()
@click.pass_context
@click.version_option()
@click.option('--api-key', help='API key to login to the platform.', required=False)
@click.option('--proxy', '-X', help='Proxy to route HTTPS requests through.')
@click.option('--tlsverify/--no-tlsverify', help='Control TLS server certificate verification.', default=True)
@click.option('-d', '--debug/--no-debug', help='Enable debug mode', default=False)
@click.option('-v', '--verbose/--no-verbose', help='Enable verbose mode', default=False)
def rootcli(ctx: click.core.Context, proxy: Optional[str] = None, tlsverify: Optional[bool] = True, debug: bool = False,
            verbose: bool = False, api_key: str = None) -> None:
    """Ostorlab is an open-source project to help automate security testing.\n
    Ostorlab standardizes interoperability between tools in a consistent, scalable, and performant way."""
    ctx.obj = {}
    ctx.obj['proxy'] = proxy
    ctx.obj['tlsverify'] = tlsverify
    ctx.obj['api_key'] = api_key
    # Configuration is a singleton class. One initiated with the provided API key, others will make use of it.
    conf_manager = configuration_manager.ConfigurationManager()
    conf_manager.api_key = api_key
    ctx.obj['config_manager'] = conf_manager

    if verbose is True:
        loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]
        for l in loggers:
            l.setLevel(logging.INFO)
    if debug is True:
        loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]
        for l in loggers:
            l.setLevel(logging.DEBUG)



"""Display the agent install progress."""
from typing import Dict, Iterator

from rich import progress

from ostorlab.cli import console as cli_console


console = cli_console.Console()


class AgentInstallProgress(progress.Progress):
    """Class reponsible for displaying the progress of the agent installation."""

    def __init__(self):
        super().__init__('[progress.description]{task.description}',
                                   progress.BarColumn(bar_width=None),
                                   '[progress.percentage]{task.percentage:>3.1f}%',
                                   '•',
                                   progress.DownloadColumn(),
                                   '•',
                                   progress.TransferSpeedColumn(),
                                   '•',
                                   progress.TimeRemainingColumn())

    def display(self, logs_generator: Iterator[Dict]) -> None:
        """Display the progress of the agent install command.

        Args:
            logs_generator: generator of the agent install command logs.
        """
        download_tasks = {}
        extract_tasks = {}

        with self as prg:
            for log in logs_generator:
                if 'Downloading' in log.get('status', []):
                    task_id = log['id']
                    if task_id not in download_tasks:
                        task = prg.add_task(f'[red]Download : {task_id}',
                                            total=log['progressDetail']['total'])
                        download_tasks[task_id] = task
                    task = download_tasks[task_id]
                    current_progress = log['progressDetail']['current']
                    prg.update(task, completed=current_progress)

                elif 'Download complete' in log.get('status', []):
                    task_id = log['id']
                    if task_id in download_tasks:
                        task = download_tasks[task_id]
                        prg.update(task, total=100.0)

                elif 'Extracting' in log.get('status', []):
                    task_id = log['id']
                    if task_id not in extract_tasks:
                        task = prg.add_task(f'[red]Extract : {task_id}',
                                            total=log['progressDetail']['total'])
                        extract_tasks[task_id] = task
                    task = extract_tasks[task_id]
                    current_progress = log['progressDetail']['current']
                    prg.update(task, completed=current_progress)

        console.success('Installation successful')



"""MQ Mixin.

Defintion of the main methods to publish and consume MQ messages by the agents.
"""

import asyncio
import concurrent.futures
import logging
from typing import List, Optional

import aio_pika

logger = logging.getLogger(__name__)


class AgentMQMixin:
    """MQ Mixin class used to initialize the channel, send messages and process them."""

    def __init__(self,
                 name: str,
                 keys: List[str],
                 url: str,
                 topic: str,
                 max_priority: Optional[int] = None,
                 loop: Optional[asyncio.AbstractEventLoop] = None):
        """Initialize the MQ parameters, the channel pools and the executors to process the messages.
        Args:
            name: Name of the queue.
            keys: Selectors that the queue listens to.
            url: string for the MQ url might contain ssl parameters e.g.`amqps://user:pass@host//`
            topic: string with exchange name
            max_priority: Optional To declare a priority queue, it is a positive int [1, 255],
             indicating the max priority the queue supports.
        """
        self._name = name
        self._keys = keys
        self._queue_name = f'{self._name}_queue'
        self._url = url
        self._topic = topic
        self._loop = loop or asyncio.get_event_loop()
        self._max_priority = max_priority
        self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=3)
        self._queue: Optional[aio_pika.Queue] = None
        self._connection_pool: aio_pika.pool.Pool[aio_pika.Connection] = aio_pika.pool.Pool(self._get_connection,
                                                                                            max_size=32,
                                                                                            loop=self._loop)
        self._channel_pool: aio_pika.pool.Pool[aio_pika.Channel] = aio_pika.pool.Pool(self._get_channel, max_size=64,
                                                                                      loop=self._loop)

    async def _get_connection(self) -> aio_pika.Connection:
        return await aio_pika.connect_robust(url=self._url, loop=self._loop)

    async def _get_channel(self) -> aio_pika.Channel:
        async with self._connection_pool.acquire() as connection:
            channel: aio_pika.Channel = await connection.channel()
            return channel

    async def _get_exchange(self, channel: aio_pika.Channel) -> aio_pika.Exchange:
        return await channel.declare_exchange(self._topic,
                                              type=aio_pika.ExchangeType.TOPIC,
                                              arguments={'x-max-length': 10000, 'x-overflow': 'reject-publish'})

    async def mq_init(self, delete_queue_first: bool = False) -> None:
        """Declares the queue to start preparing for receiving messages.
        Args:
            delete_queue_first: Used for testing purposes. To delete pending queues first.
        """
        logger.info('Connecting to %s', self._url)
        connection = await self._get_connection()
        async with connection:
            channel = await connection.channel()
            await self._declare_mq_queue(channel, delete_queue_first)

    async def mq_run(self, delete_queue_first: bool = False) -> None:
        """Use a channel to declare the queue, set the listener on the selectors and consume the received messaged.
        Args:
            delete_queue_first: Used for testing purposes. To delete pending queues first.
        """
        connection = await self._get_connection()
        channel = await connection.channel()
        await self._declare_mq_queue(channel, delete_queue_first)

        if self._queue is None:
            raise ValueError('queue not created')

        await self._queue.consume(self._mq_process_message, no_ack=False)

    async def _declare_mq_queue(self, channel: aio_pika.Channel, delete_queue_first: bool = False) -> None:
        """Declare the MQ queue on a given channel.
        The queue is durable, re-declaring the queue will return the same queue
        Args:
            channel: the MQ channel to use for the queues
            delete_queue_first: Used for testing purposes. To delete pending queues first.
        """
        await channel.set_qos(prefetch_count=1)
        exchange = await self._get_exchange(channel)
        if delete_queue_first:
            await channel.queue_delete(self._queue_name)

        if self._max_priority is not None:
            self._queue = await channel.declare_queue(self._queue_name, auto_delete=False, durable=True,
                                                      arguments={'x-max-priority': self._max_priority})
        else:
            self._queue = await channel.declare_queue(self._queue_name, auto_delete=False, durable=True)

        if self._queue is None:
            raise ValueError('queue not declared.')

        for k in self._keys:
            await self._queue.bind(exchange, k)

    async def _mq_process_message(self, message: aio_pika.IncomingMessage) -> None:
        """Consumes the MQ messages and calls the process message callback."""
        logger.debug('incoming pika message received')
        async with message.process(requeue=True, reject_on_redelivered=True):
            await self._loop.run_in_executor(self._executor,
                                             self.process_message,
                                             message.routing_key,
                                             message.body)

    def process_message(self, selector: str, message: bytes) -> None:
        """Callback to implement to process the MQ messages received."""
        raise NotImplementedError()

    async def async_mq_send_message(self, key: str, message: bytes, message_priority: Optional[int] = None) -> None:
        """Async Send the message to the provided routing key and its priority.
        Args:
            message: Message to send
            key: Selector where to send the message.
            message_priority: the priority of the message. Default is 0
        """
        logger.debug('sending %s to %s', message, key)
        async with self._channel_pool.acquire() as channel:
            exchange = await self._get_exchange(channel)
            pika_message = aio_pika.Message(body=message, priority=message_priority or 0)
            await exchange.publish(routing_key=key, message=pika_message)

    def mq_send_message(self, key: str, message: bytes, message_priority: Optional[int] = None) -> None:
        """the method sends the message to the selected key with the defined priority in async mode .
        Args:
            keys: Selectors that the queue listens to.
            message: Message to send .
            message_priority: the priority to use for the message default is 0.
        """
        logger.debug('sending %s to %s', message, key)
        if not self._loop.is_running():
            self._loop.run_until_complete(self.async_mq_send_message(key, message, message_priority))
        else:
            self._loop.create_task(self.async_mq_send_message(key, message, message_priority))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ostorlab/agent/message/proto/common/x509/x509.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='ostorlab/agent/message/proto/common/x509/x509.proto',
  package='ostorlab.agent.message.proto.common.x509',
  syntax='proto2',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n3ostorlab/agent/message/proto/common/x509/x509.proto\x12(ostorlab.agent.message.proto.common.x509\"N\n\x07X509Org\x12\t\n\x01\x43\x18\x01 \x01(\t\x12\t\n\x01O\x18\x02 \x01(\t\x12\n\n\x02OU\x18\x03 \x01(\t\x12\n\n\x02\x43N\x18\x04 \x01(\t\x12\t\n\x01L\x18\x05 \x01(\t\x12\n\n\x02ST\x18\x06 \x01(\t\"1\n\x08Validity\x12\x12\n\nnot_before\x18\x01 \x01(\x04\x12\x11\n\tnot_after\x18\x02 \x01(\x04\"K\n\x0fX509Fingerprint\x12\x1d\n\x15\x66ingerprint_algorithm\x18\x01 \x01(\t\x12\x19\n\x11\x66ingerprint_value\x18\x02 \x01(\t\"=\n\rX509Extension\x12\x0b\n\x03oid\x18\x01 \x01(\t\x12\x10\n\x08\x63ritical\x18\x02 \x01(\x08\x12\r\n\x05value\x18\x03 \x01(\t\"G\n\rX509PublicKey\x12\x1c\n\x14public_key_algorithm\x18\x01 \x01(\t\x12\x18\n\x10public_key_value\x18\x02 \x01(\t\"E\n\rX509Signature\x12\x1b\n\x13signature_algorithm\x18\x01 \x01(\t\x12\x17\n\x0fsignature_value\x18\x02 \x01(\t\"\xc5\x04\n\x08X509Cert\x12\x16\n\x0eversion_number\x18\x01 \x01(\r\x12\x15\n\rserial_number\x18\x02 \x01(\t\x12\x41\n\x06issuer\x18\x03 \x01(\x0b\x32\x31.ostorlab.agent.message.proto.common.x509.X509Org\x12\x44\n\x08validity\x18\x04 \x01(\x0b\x32\x32.ostorlab.agent.message.proto.common.x509.Validity\x12\x42\n\x07subject\x18\x05 \x01(\x0b\x32\x31.ostorlab.agent.message.proto.common.x509.X509Org\x12K\n\npublic_key\x18\x06 \x01(\x0b\x32\x37.ostorlab.agent.message.proto.common.x509.X509PublicKey\x12J\n\tsignature\x18\x07 \x01(\x0b\x32\x37.ostorlab.agent.message.proto.common.x509.X509Signature\x12W\n\x14\x63omputed_fingerprint\x18\x08 \x03(\x0b\x32\x39.ostorlab.agent.message.proto.common.x509.X509Fingerprint\x12K\n\nextensions\x18\t \x03(\x0b\x32\x37.ostorlab.agent.message.proto.common.x509.X509Extension'
)




_X509ORG = _descriptor.Descriptor(
  name='X509Org',
  full_name='ostorlab.agent.message.proto.common.x509.X509Org',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='C', full_name='ostorlab.agent.message.proto.common.x509.X509Org.C', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='O', full_name='ostorlab.agent.message.proto.common.x509.X509Org.O', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='OU', full_name='ostorlab.agent.message.proto.common.x509.X509Org.OU', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='CN', full_name='ostorlab.agent.message.proto.common.x509.X509Org.CN', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='L', full_name='ostorlab.agent.message.proto.common.x509.X509Org.L', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ST', full_name='ostorlab.agent.message.proto.common.x509.X509Org.ST', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=97,
  serialized_end=175,
)


_VALIDITY = _descriptor.Descriptor(
  name='Validity',
  full_name='ostorlab.agent.message.proto.common.x509.Validity',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='not_before', full_name='ostorlab.agent.message.proto.common.x509.Validity.not_before', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='not_after', full_name='ostorlab.agent.message.proto.common.x509.Validity.not_after', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=177,
  serialized_end=226,
)


_X509FINGERPRINT = _descriptor.Descriptor(
  name='X509Fingerprint',
  full_name='ostorlab.agent.message.proto.common.x509.X509Fingerprint',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='fingerprint_algorithm', full_name='ostorlab.agent.message.proto.common.x509.X509Fingerprint.fingerprint_algorithm', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fingerprint_value', full_name='ostorlab.agent.message.proto.common.x509.X509Fingerprint.fingerprint_value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=228,
  serialized_end=303,
)


_X509EXTENSION = _descriptor.Descriptor(
  name='X509Extension',
  full_name='ostorlab.agent.message.proto.common.x509.X509Extension',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='oid', full_name='ostorlab.agent.message.proto.common.x509.X509Extension.oid', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='critical', full_name='ostorlab.agent.message.proto.common.x509.X509Extension.critical', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='ostorlab.agent.message.proto.common.x509.X509Extension.value', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=305,
  serialized_end=366,
)


_X509PUBLICKEY = _descriptor.Descriptor(
  name='X509PublicKey',
  full_name='ostorlab.agent.message.proto.common.x509.X509PublicKey',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='public_key_algorithm', full_name='ostorlab.agent.message.proto.common.x509.X509PublicKey.public_key_algorithm', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='public_key_value', full_name='ostorlab.agent.message.proto.common.x509.X509PublicKey.public_key_value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=368,
  serialized_end=439,
)


_X509SIGNATURE = _descriptor.Descriptor(
  name='X509Signature',
  full_name='ostorlab.agent.message.proto.common.x509.X509Signature',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='signature_algorithm', full_name='ostorlab.agent.message.proto.common.x509.X509Signature.signature_algorithm', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='signature_value', full_name='ostorlab.agent.message.proto.common.x509.X509Signature.signature_value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=441,
  serialized_end=510,
)


_X509CERT = _descriptor.Descriptor(
  name='X509Cert',
  full_name='ostorlab.agent.message.proto.common.x509.X509Cert',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='version_number', full_name='ostorlab.agent.message.proto.common.x509.X509Cert.version_number', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='serial_number', full_name='ostorlab.agent.message.proto.common.x509.X509Cert.serial_number', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='issuer', full_name='ostorlab.agent.message.proto.common.x509.X509Cert.issuer', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='validity', full_name='ostorlab.agent.message.proto.common.x509.X509Cert.validity', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='subject', full_name='ostorlab.agent.message.proto.common.x509.X509Cert.subject', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='public_key', full_name='ostorlab.agent.message.proto.common.x509.X509Cert.public_key', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='signature', full_name='ostorlab.agent.message.proto.common.x509.X509Cert.signature', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='computed_fingerprint', full_name='ostorlab.agent.message.proto.common.x509.X509Cert.computed_fingerprint', index=7,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='extensions', full_name='ostorlab.agent.message.proto.common.x509.X509Cert.extensions', index=8,
      number=9, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=513,
  serialized_end=1094,
)

_X509CERT.fields_by_name['issuer'].message_type = _X509ORG
_X509CERT.fields_by_name['validity'].message_type = _VALIDITY
_X509CERT.fields_by_name['subject'].message_type = _X509ORG
_X509CERT.fields_by_name['public_key'].message_type = _X509PUBLICKEY
_X509CERT.fields_by_name['signature'].message_type = _X509SIGNATURE
_X509CERT.fields_by_name['computed_fingerprint'].message_type = _X509FINGERPRINT
_X509CERT.fields_by_name['extensions'].message_type = _X509EXTENSION
DESCRIPTOR.message_types_by_name['X509Org'] = _X509ORG
DESCRIPTOR.message_types_by_name['Validity'] = _VALIDITY
DESCRIPTOR.message_types_by_name['X509Fingerprint'] = _X509FINGERPRINT
DESCRIPTOR.message_types_by_name['X509Extension'] = _X509EXTENSION
DESCRIPTOR.message_types_by_name['X509PublicKey'] = _X509PUBLICKEY
DESCRIPTOR.message_types_by_name['X509Signature'] = _X509SIGNATURE
DESCRIPTOR.message_types_by_name['X509Cert'] = _X509CERT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

X509Org = _reflection.GeneratedProtocolMessageType('X509Org', (_message.Message,), {
  'DESCRIPTOR' : _X509ORG,
  '__module__' : 'ostorlab.agent.message.proto.common.x509.x509_pb2'
  # @@protoc_insertion_point(class_scope:ostorlab.agent.message.proto.common.x509.X509Org)
  })
_sym_db.RegisterMessage(X509Org)

Validity = _reflection.GeneratedProtocolMessageType('Validity', (_message.Message,), {
  'DESCRIPTOR' : _VALIDITY,
  '__module__' : 'ostorlab.agent.message.proto.common.x509.x509_pb2'
  # @@protoc_insertion_point(class_scope:ostorlab.agent.message.proto.common.x509.Validity)
  })
_sym_db.RegisterMessage(Validity)

X509Fingerprint = _reflection.GeneratedProtocolMessageType('X509Fingerprint', (_message.Message,), {
  'DESCRIPTOR' : _X509FINGERPRINT,
  '__module__' : 'ostorlab.agent.message.proto.common.x509.x509_pb2'
  # @@protoc_insertion_point(class_scope:ostorlab.agent.message.proto.common.x509.X509Fingerprint)
  })
_sym_db.RegisterMessage(X509Fingerprint)

X509Extension = _reflection.GeneratedProtocolMessageType('X509Extension', (_message.Message,), {
  'DESCRIPTOR' : _X509EXTENSION,
  '__module__' : 'ostorlab.agent.message.proto.common.x509.x509_pb2'
  # @@protoc_insertion_point(class_scope:ostorlab.agent.message.proto.common.x509.X509Extension)
  })
_sym_db.RegisterMessage(X509Extension)

X509PublicKey = _reflection.GeneratedProtocolMessageType('X509PublicKey', (_message.Message,), {
  'DESCRIPTOR' : _X509PUBLICKEY,
  '__module__' : 'ostorlab.agent.message.proto.common.x509.x509_pb2'
  # @@protoc_insertion_point(class_scope:ostorlab.agent.message.proto.common.x509.X509PublicKey)
  })
_sym_db.RegisterMessage(X509PublicKey)

X509Signature = _reflection.GeneratedProtocolMessageType('X509Signature', (_message.Message,), {
  'DESCRIPTOR' : _X509SIGNATURE,
  '__module__' : 'ostorlab.agent.message.proto.common.x509.x509_pb2'
  # @@protoc_insertion_point(class_scope:ostorlab.agent.message.proto.common.x509.X509Signature)
  })
_sym_db.RegisterMessage(X509Signature)

X509Cert = _reflection.GeneratedProtocolMessageType('X509Cert', (_message.Message,), {
  'DESCRIPTOR' : _X509CERT,
  '__module__' : 'ostorlab.agent.message.proto.common.x509.x509_pb2'
  # @@protoc_insertion_point(class_scope:ostorlab.agent.message.proto.common.x509.X509Cert)
  })
_sym_db.RegisterMessage(X509Cert)


# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ostorlab/agent/message/proto/v3/asset/file/file.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n5ostorlab/agent/message/proto/v3/asset/file/file.proto\x12*ostorlab.agent.message.proto.v3.asset.file\"(\n\x07Message\x12\x0f\n\x07\x63ontent\x18\x01 \x01(\x0c\x12\x0c\n\x04path\x18\x02 \x01(\t')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ostorlab.agent.message.proto.v3.asset.file.file_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _MESSAGE._serialized_start=101
  _MESSAGE._serialized_end=141
# @@protoc_insertion_point(module_scope)



"""Common logic for Mobile Assets of type .APK package file AAB package file and IPA .
This module takes care of preparing the application file and calling the create mobile scan API.
"""
import io
import click
import itertools

from typing import List

from ostorlab.cli.ci_scan.run import run
from ostorlab.apis.runners import authenticated_runner
from ostorlab.apis.runners import runner as base_runner
from ostorlab.apis import scan_create as scan_create_api
from ostorlab.apis import test_credentials_create as test_credentials_create_api


def _prepare_test_credentials(ctx: click.core.Context) -> List[test_credentials_create_api.TestCredential]:
    test_credentials_login = ctx.obj['test_credentials']['test_credentials_login']
    test_credentials_password = ctx.obj['test_credentials']['test_credentials_password']
    test_credentials_url = ctx.obj['test_credentials']['test_credentials_url']
    test_credentials_role = ctx.obj['test_credentials']['test_credentials_role']
    test_credentials_name = ctx.obj['test_credentials']['test_credentials_name']
    test_credentials_value = ctx.obj['test_credentials']['test_credentials_value']

    credentials = []
    for login, password, role, url in itertools.zip_longest(
            test_credentials_login,
            test_credentials_password,
            test_credentials_role,
            test_credentials_url):
        credentials.append(test_credentials_create_api.TestCredentialLogin(
            login=login,
            password=password,
            role=role,
            url=url
        ))

    if test_credentials_name:
        credentials.append(test_credentials_create_api.TestCredentialCustom(
            values=dict(itertools.zip_longest(test_credentials_name, test_credentials_value))))

    return credentials


def run_mobile_scan(ctx: click.core.Context, file: io.FileIO, asset_type: scan_create_api.MobileAssetType) -> None:
    """Create scan for mobile application package file."""
    ci_logger = ctx.obj['ci_logger']
    if ctx.obj.get('api_key'):
        scan_profile = ctx.obj['scan_profile']
        title = ctx.obj['title']
        break_on_risk_rating = ctx.obj['break_on_risk_rating']
        max_wait_minutes = ctx.obj['max_wait_minutes']
        runner = authenticated_runner.AuthenticatedAPIRunner(api_key=ctx.obj.get('api_key'))
        try:

            test_credentials = _prepare_test_credentials(ctx)
            if test_credentials:
                credential_ids = _create_test_credentials(test_credentials, runner)
                ci_logger.info(f'Created test credentials {", ".join(str(t) for t in test_credentials)} successfully')
            else:
                credential_ids = []

            ci_logger.info(f'creating scan `{title}` with profile `{scan_profile}` for `{asset_type}`')
            scan_id = _create_scan(title, scan_profile, asset_type, file, credential_ids, runner)

            ci_logger.output(name='scan_id', value=scan_id)
            ci_logger.info(f'Scan created with id {scan_id}.')

            if break_on_risk_rating is not None:
                run.apply_break_scan_risk_rating(break_on_risk_rating, scan_id, max_wait_minutes, runner, ci_logger)

        except base_runner.ResponseError as e:
            ci_logger.error(f'Could not start the scan. {e}')
            raise click.exceptions.Exit(2) from e
        except TimeoutError:
            ci_logger.error('The scan is still running.')
            raise click.exceptions.Exit(2) from None
    else:
        ci_logger.error('API key not not provided.')
        raise click.exceptions.Exit(2) from None


def _create_scan(title, scan_profile, asset_type, file, credential_ids, runner):
    scan_result = runner.execute(
        scan_create_api.CreateMobileScanAPIRequest(title=title,
                                                   asset_type=asset_type,
                                                   scan_profile=scan_profile,
                                                   application=file,
                                                   test_credential_ids=credential_ids
                                                   ))
    scan_id = scan_result.get('data').get('createMobileScan').get('scan').get('id')
    return scan_id


def _create_test_credentials(test_credentials, runner):
    credential_ids = []
    for test_credential in test_credentials:
        test_credential_result = runner.execute(
            test_credentials_create_api.CreateTestCredentialAPIRequest(test_credential=test_credential))
        credential_ids.append(
            test_credential_result.get('data').get('createTestCredentials').get('testCredentials').get('id'))
    return credential_ids



"""Ostorlab cli package."""
from ostorlab.cli import scan
from ostorlab.cli import auth
from ostorlab.cli import agent
from ostorlab.cli import vulnz
from ostorlab.cli import agentgroup
from ostorlab.cli import ci_scan



"""Jaeger service to collect Agent traces."""
import logging
from typing import Dict, Optional

import docker
import tenacity
from docker import errors
from docker import types
from docker.models import services

logger = logging.getLogger(__name__)

JAEGER_IMAGE = 'jaegertracing/all-in-one:latest'
DEFAULT_EXPOSED_PORTS = {5775: 5775, 6831: 6831, 6832: 6832, 5778: 5778, 16686: 16686, 14268: 14268, 9411: 9411}
DEFAULT_JAEGER_PORT = 6831
DEFAULT_JAEGER_UI_PORT = 16686

class LocalJaeger:
    """Jaeger service spawned a docker swarm service."""

    def __init__(self,
                 name: str,
                 network: str,
                 exposed_ports: Dict[int, int] = None,
                 image: str = JAEGER_IMAGE) -> None:
        """Initialize the Jaeger service parameters.
        Args:
            name: Name of the service.
            network: Network used for the Docker Jaeger service.
            exposed_ports: The list of Jaeger service exposed ports
            image: Jaeger Docker image
        """
        self._name = name
        self._docker_client = docker.from_env()
        # images
        self._jaeger_image = image
        self._network = network
        self._jaeger_host = f'jaeger_{self._name}'
        # service
        self._jaeger_service = None
        # exposed_port
        self._exposed_ports = exposed_ports or DEFAULT_EXPOSED_PORTS

    @property
    def url(self) -> str:
        """URL to connect to the local Jaeger instance."""
        return f'jaeger://{self._jaeger_host}:{DEFAULT_JAEGER_PORT}'

    @property
    def management_interface_ui(self) -> str:
        """URL to management user interface of the Jaeger instance."""
        return f'http://127.0.0.1:{DEFAULT_JAEGER_UI_PORT}'

    @property
    def service(self):
        return self._jaeger_service

    def start(self) -> None:
        """Start local Jaeger instance."""
        self._create_network()
        self._jaeger_service = self._start_jaeger()

        if self._jaeger_service is not None and not self._is_service_healthy():
            logger.error('Jaeger container for service %s is not ready', self._jaeger_service.id)
            return

    def stop(self):
        """Stop the local Jaeger instance."""
        for service in self._docker_client.services.list():
            universe = service.attrs['Spec']['Labels'].get('ostorlab.universe')
            if universe is not None and service.name.startswith('jaeger_') and self._name in universe:
                service.remove()

    def _create_network(self):
        if any(network.name == self._network for network in self._docker_client.networks.list()):
            logger.warning('network already exists.')
        else:
            logger.info('creating private network %s', self._network)
            return self._docker_client.networks.create(
                name=self._network,
                driver='overlay',
                attachable=True,
                labels={'ostorlab.universe': self._name},
                check_duplicate=True
            )

    def _start_jaeger(self) -> Optional[services.Service]:
        try:
            logger.info('starting Jaeger')
            endpoint_spec = types.services.EndpointSpec(mode='vip', ports=self._exposed_ports)
            service_mode = types.services.ServiceMode('replicated', replicas=1)
            return self._docker_client.services.create(
                image=self._jaeger_image,
                networks=[self._network],
                name=self._jaeger_host,
                restart_policy=types.RestartPolicy(condition='any'),
                mode=service_mode,
                labels={'ostorlab.universe': self._name, 'ostorlab.jaeger': ''},
                endpoint_spec=endpoint_spec)
        except docker.errors.APIError as e:
            error_message = f'Jaeger service could not be started. Reason: {e}.'
            logger.error(error_message)
            return

    @tenacity.retry(stop=tenacity.stop_after_attempt(20),
                    wait=tenacity.wait_exponential(multiplier=1, max=12),
                    # return last value and don't raise RetryError exception.
                    retry_error_callback=lambda lv: lv.result(),
                    retry=tenacity.retry_if_result(lambda v: v is False))
    def _is_service_healthy(self) -> bool:
        logger.info('checking service %s', self._jaeger_service.name)
        return self.is_healthy

    @property
    def is_healthy(self) -> bool:
        try:
            return self._jaeger_service is not None and \
                   len([task for task in self._jaeger_service.tasks() if task['Status']['State'] == 'running']) == 1
        except errors.DockerException:
            return False



"""Handles all authenticated API calls and behind the scenes operations such as authentication, validation, etc.

    Typical usage example:

    authenticated_runner = AuthenticatedAPIRunner(username, password, token_duration)
    authenticated_runner.authenticate()
"""
import datetime
import logging
from typing import Dict, Optional, Any

import click
import requests

from ostorlab.apis import create_api_key
from ostorlab.apis import request as api_request
from ostorlab.apis.runners import login_runner
from ostorlab.apis.runners import runner
from ostorlab.cli import console as cli_console

AUTHENTICATED_GRAPHQL_ENDPOINT = 'https://api.ostorlab.co/apis/graphql'

logger = logging.getLogger(__name__)
console = cli_console.Console()


class AuthenticationError(runner.Error):
    """Authentication Error."""


class AuthenticatedAPIRunner(runner.APIRunner):
    """Handles all authenticated API calls and behind the scenes operations such as authentication,
       validation, etc.
    """

    def __init__(self,
                 username: Optional[str] = None,
                 password: Optional[str] = None,
                 token_duration: Optional[datetime.datetime] = None,
                 proxy: Optional[str] = None,
                 verify: bool = True,
                 api_key: Optional[str] = None
                 ):
        """Constructs all the necessary attributes for the object.

        Args:
            username: the username (email) used to log in to use the token based authentication.
            password: the password used to log in to use the token based authentication.
            token_duration: The duration for which the token is valid
            (Can be in minutes, hours, days, or a combination of any two or all three).
            proxy: The proxy through which a request is made. Defaults to None.
            verify: Whether to verify the TLS certificate. Defaults to True.
            api_key: Use API KEY based authentication. Used if token is not defined.
        """
        super().__init__(proxy, verify)
        self._username = username
        self._password = password
        self._token_duration = token_duration
        self._api_key = api_key or self._configuration_manager.api_key
        self._token: Optional[str] = None
        self._otp_token: Optional[str] = None

    @property
    def endpoint(self) -> str:
        """API endpoint."""
        return AUTHENTICATED_GRAPHQL_ENDPOINT

    def authenticate(self) -> None:
        """Authenticates the user.

        Raises:
            AuthenticationError: If user credentials are not valid.
        """
        with console.status('Logging into your account'):
            login_api_runner = login_runner.LoginAPIRunner(username=self._username,
                                                           password=self._password,
                                                           otp_token=self._otp_token,
                                                           proxy=self._proxy,
                                                           verify=self._verify)
            response = login_api_runner.login_user()

        if response.status_code != 200:
            field_errors = response.json().get('non_field_errors')
            if field_errors is not None and field_errors[0] == 'Must include \"otp_token\"':
                self._otp_token = click.prompt(
                    'Please enter the OTP code from your authenticator app')
                self.authenticate()
            else:
                logger.debug('response %s', response.content)
                raise AuthenticationError(response.status_code)
        else:
            self._token = response.json().get('token')
            with console.status('Generating API key'):
                api_key_response = self.execute(
                    create_api_key.CreateAPIKeyAPIRequest(self._token_duration))
                console.success('API key generated')

            api_data = api_key_response['data']['createApiKey']['apiKey']
            secret_key = api_data['secretKey']
            api_key_id = api_data['apiKey']['id']
            expiry_date = api_data['apiKey']['expiryDate']

            self._api_key = secret_key
            with console.status('Persisting API key'):
                self._configuration_manager.set_api_data(
                    secret_key, api_key_id, expiry_date)
                console.success('API key persisted')
            self._token = None
            console.success(':white_check_mark: Authentication successful')

    def unauthenticate(self) -> None:
        self._api_key = None

    def execute(self, request: api_request.APIRequest) -> Dict[str, Any]:
        """Executes a request using the Authenticated GraphQL API.

        Args:
            request: The request to be executed

        Raises:
            ResponseError: When the API returns an error

        Returns:
            The API response
        """
        if self._token is not None:
            headers = {'Authorization': f'Token {self._token}'}
        elif self._api_key is not None:
            headers = {'X-Api-Key': f'{self._api_key}'}
        else:
            headers = None
            console.warning('No authentication credentials were provided.')

        response = self._sent_request(request, headers)
        if response.status_code != 200:
            logger.debug('Response status code is %s: %s', response.status_code, response.content)
            raise runner.ResponseError(
                f'Response status code is {response.status_code}: {response.content.decode(errors="ignore")}')
        data: Dict[str, Any] = response.json()
        errors = data.get('errors')
        if errors is not None and isinstance(errors, list):
            error = errors[0].get('message')
            raise runner.ResponseError(f'Response errors: {error}')
        else:
            return data

    def _sent_request(self, request: api_request.APIRequest,
                      headers: Optional[Dict[str, str]] = None) -> requests.Response:
        """Sends an API request."""
        if self._proxy is not None:
            proxy = {
                'https': self._proxy
            }
        else:
            proxy = None

        return requests.post(self.endpoint, data=request.data, files=request.files, headers=headers,
                             proxies=proxy, verify=self._verify, timeout=runner.REQUEST_TIMEOUT)



"""Android .AAB asset."""
import dataclasses
from typing import Optional

from ostorlab.assets import asset


@asset.selector('v3.asset.file.android.aab')
@dataclasses.dataclass
class AndroidAab(asset.Asset):
    """Android .AAB target asset."""

    def __init__(self, content: bytes, path: Optional[str] = None):
        self.content = content
        self.path = path

    def __str__(self) -> str:
        return f'Android AAB({self.path})'

    @property
    def proto_field(self) -> str:
        return 'android_aab'



"""Create test credentials API."""
import abc
import dataclasses
import json
from typing import Dict, Optional, Any

from . import request


class TestCredential(abc.ABC):
    """Base abstract test credentials."""

    @abc.abstractmethod
    def to_variables(self) -> Dict[str, Any]:
        """Generate query variables."""
        pass


@dataclasses.dataclass
class TestCredentialLogin(TestCredential):
    """Loging password test credentials with optional role and url fields."""
    login: str
    password: str
    role: Optional[str] = None
    url: Optional[str] = None

    def to_variables(self) -> Dict[str, Any]:
        """Generate query variables."""
        return {
            'testCredentials': {
                'loginPassword': {
                    'login': self.login,
                    'password': self.password,
                    'role': self.role,
                    'url': self.url
                }
            }
        }


@dataclasses.dataclass
class TestCredentialCustom(TestCredential):
    """Custom test credentials with variable number of a pair of name and value."""
    values: Dict[str, str]

    def to_variables(self) -> Dict[str, Any]:
        """Generate query variables."""
        return {
            'testCredentials': {
                'custom': {
                    'credentials': [
                        {
                            'name': n, 'value': v
                        } for (n, v) in self.values.items()
                    ]
                }
            }
        }


class CreateTestCredentialAPIRequest(request.APIRequest):
    """Create mobile scan API from a file."""

    def __init__(self, test_credential: TestCredential):
        self._test_credential = test_credential

    @property
    def query(self) -> Optional[str]:
        """Defines the query to create test credentials.

        Returns:
            The query to create a test credential.
        """

        return """
mutation TestCredentials($testCredentials: TestCredentialsInput!) {
  createTestCredentials(testCredentials: $testCredentials) {
    testCredentials {
      ... on CustomTestCredentials {
        id
      }
      ... on LoginPasswordTestCredentials {
        id
      }
    }
  }
}
        """

    @property
    def data(self) -> Optional[Dict]:
        """Sets the query and variables to create test credentials.

        Returns:
            The query and variables to create test credentials.
         """
        data = {
            'query': self.query,
            'variables': json.dumps(self._test_credential.to_variables())
        }
        return data



"""IP address asset."""
import dataclasses
import ipaddress
from typing import Optional

from ostorlab.assets import asset


@dataclasses.dataclass
@asset.selector('v3.asset.ip')
class IP(asset.Asset):
    """IP Address target asset."""

    def __init__(self, host: str, version: Optional[int] = None, mask: Optional[str] = None):
        self.host = host
        self.version = version or ipaddress.ip_address(self.host).version
        self.mask = mask

    def __str__(self) -> str:
        return f'{self.host}/{self.mask}'



"""Agent and Agent Group loader helper methods that handles schema validation and safe loading."""
import io
import pathlib

import ruamel.yaml

from ostorlab.agent.schema import validator

from typing import Dict, Any

AGENT_SPEC_PATH = pathlib.Path(__file__).parent / 'agent_schema.json'
AGENT_GROUP_SPEC_PATH = pathlib.Path(__file__).parent / 'agent_group_schema.json'


def _load_spec_yaml(file: io.TextIOWrapper, spec: pathlib.Path) -> Dict[str, Any]:
    """Loads file based on spec"""
    with open(spec, 'r', encoding='utf8') as agent_spec:
        yaml_def_validator = validator.Validator(agent_spec)
        yaml_def_validator.validate(file)
        file.seek(0)
        yaml = ruamel.yaml.YAML(typ='safe')
        agent_def: Dict[str, Any] = yaml.load(file)
        return agent_def


def load_agent_yaml(file: io.TextIOWrapper) -> Dict[str, Any]:
    """Loads and validates agent yaml definition file.

    Args:
        file: Yaml definition file.

    Returns:
        Parsed yaml dict.
    """

    spec = AGENT_SPEC_PATH
    return _load_spec_yaml(file, spec)


def load_agent_group_yaml(file: io.TextIOWrapper) -> Dict[str, Any]:
    """Loads and validates agent gorup yaml definition file.

    Args:
        file: Yaml definition file.

    Returns:
        Parsed yaml dict.
    """
    spec = AGENT_GROUP_SPEC_PATH
    return _load_spec_yaml(file, spec)



"""Asset of type .AAB package file.
This module takes care of preparing a file of type .aab before injecting it to the runtime instance."""
import io
import logging
from typing import List

import click

from ostorlab.assets import android_aab as android_aab_asset
from ostorlab.cli.scan.run import run

logger = logging.getLogger(__name__)


@run.run.command()
@click.argument('files', type=click.File(mode='rb'), nargs=-1, required=True)
@click.pass_context
def android_aab(ctx: click.core.Context, files: List[io.FileIO]) -> None:
    """Run scan for android .AAB package file."""
    runtime = ctx.obj['runtime']
    assets = []
    for f in files:
        assets.append(android_aab_asset.AndroidAab(content=f.read(), path=str(f.name)))
    logger.debug('scanning assets %s', [str(asset) for asset in assets])
    runtime.scan(title=ctx.obj['title'], agent_group_definition=ctx.obj['agent_group_definition'], assets=assets)



"""Link asset."""
import dataclasses

from ostorlab.assets import asset


@dataclasses.dataclass
@asset.selector('v3.asset.link')
class Link(asset.Asset):
    """Agent asset."""

    def __init__(self, url: str, method: str):
        self.url = url
        self.method = method

    def __str__(self) -> str:
        return f'Link {self.url} with method {self.method}'

    @property
    def proto_field(self) -> str:
        return 'link'



"""Version utility class to make easy to compare semantic version strings."""
import semver


class Version:
    """Version class that only accepts semantic and provides python comparison API."""

    def __init__(self, version: str) -> None:
        """Init version."""
        self._version = version
        self._semver = semver.VersionInfo.parse(version)

    def __repr__(self) -> str:
        """Version string representation."""
        return f'<Version {self._version}>'

    def __str__(self) -> str:
        """Version string"""
        return self._version

    def __lt__(self, other: object) -> bool:
        """Pythonic comparison API."""
        if isinstance(other, Version):
            return bool(self._semver.compare(other._version) < 0)
        else:
            raise ValueError()

    def __le__(self, other: object) -> bool:
        """Pythonic comparison API."""
        if isinstance(other, Version):
            return bool(self._semver.compare(other._version) <= 0)
        else:
            raise ValueError()

    def __gt__(self, other: object) -> bool:
        """Pythonic comparison API."""
        if isinstance(other, Version):
            return bool(self._semver.compare(other._version) > 0)
        else:
            raise ValueError()

    def __ge__(self, other: object) -> bool:
        """Pythonic comparison API."""
        if isinstance(other, Version):
            return bool(self._semver.compare(other._version) >= 0)
        else:
            raise ValueError()

    def __eq__(self, other: object) -> bool:
        """Pythonic comparison API."""
        if isinstance(other, Version):
            return bool(self._version == other._version)
        else:
            raise ValueError()



"""File asset."""
import dataclasses
from typing import Optional

from ostorlab.assets import asset


@dataclasses.dataclass
@asset.selector('v3.asset.file')
class File(asset.Asset):
    """File target asset."""

    def __init__(self, content: bytes, path: Optional[str] = None):
        self.content = content
        self.path = path

    def __str__(self) -> str:
        return f'File({self.path})'

    @property
    def proto_field(self) -> str:
        return 'file'



"""Vulnz dump command."""
import logging

import click
import sqlalchemy
from ostorlab.apis.runners import runner

from ostorlab.cli import console as cli_console
from ostorlab.cli import dumpers
from ostorlab.cli.vulnz import vulnz

console = cli_console.Console()

logger = logging.getLogger(__name__)


@vulnz.command(name='dump')
@click.option('--scan-id', '-s', 'scan_id', help='Id of the scan.', required=True)
@click.option('--output', '-o', help='Output file path', required=False, default='./output')
@click.option('--format', '-f', 'output_format', help='Output format',
              required=False, type=click.Choice(['jsonl', 'csv']), default='jsonl')
@click.pass_context
def dump(ctx, scan_id: int, output: str, output_format: str) -> None:
    """Dump found vulnerabilities of a scan in a specific format."""
    try:
        runtime_instance = ctx.obj['runtime']
        console.info(f'Fetching vulnerabilities for scan {scan_id}')
        if output_format == 'csv':
            dumper = dumpers.VulnzCsvDumper(output)
        else:
            dumper = dumpers.VulnzJsonDumper(output)
        runtime_instance.dump_vulnz(scan_id=scan_id, dumper=dumper)
    except FileNotFoundError as e:
        console.error(f'No such file or directory: {output}')
        raise click.exceptions.Exit(2) from e
    except (sqlalchemy.exc.OperationalError, runner.Error):
        console.error(f'scan with id {scan_id} does not exist.')



"""Auth login command."""

import logging

import click

from ostorlab.apis.runners import authenticated_runner
from ostorlab.cli.auth import auth
from ostorlab.cli import console as cli_console

logger = logging.getLogger(__name__)
console = cli_console.Console()


@auth.auth.command()
@click.option('--username', '-u', help='Ostorlab platform username.', required=True, prompt=True)
@click.option('--password', '-p', help='Ostorlab platform password.', required=True, prompt=True, hide_input=True)
@click.option('--token-duration', help='Expiration time for token (m for minutes, h for hours, and d for days).')
def login(username, password, token_duration):
    """Use this to log into your account."""
    try:
        api_runner = authenticated_runner.AuthenticatedAPIRunner(username=username, password=password,
                                           token_duration=token_duration)
        api_runner.authenticate()
    except authenticated_runner.AuthenticationError:
        console.error(
            'Authentication error, please check that your credentials are valid.')



"""Create Asset via an API Request."""
import json
from typing import Dict, Optional

from ostorlab.apis import request
from ostorlab.assets import asset as base_asset
from ostorlab.assets import android_aab
from ostorlab.assets import android_apk
from ostorlab.assets import file
from ostorlab.assets import ios_ipa
from ostorlab.assets import domain_name
from ostorlab.assets import ip
from ostorlab.assets import ipv4
from ostorlab.assets import ipv6


class CreateAssetAPIRequest(request.APIRequest):
    """Persist asset API request"""

    def __init__(self, asset: base_asset.Asset) -> None:
        """Constructs all the necessary attributes for the object."""
        self._asset = asset

    @property
    def query(self) -> Optional[str]:
        """Sets the query of the API request.

        Returns:
            The query to create the asset.
        """
        return """
            mutation CreateAsset($asset: AssetInputType!){
                createAsset(asset: $asset) {
                    asset{
                        ... on AndroidAssetType{
                            id
                        }
                        ... on IOSAssetType{
                            id
                        }
                        ... on AndroidFileAssetType{
                            id
                        }
                        ... on IOSFileAssetType{
                            id
                        }
                        ... on AndroidStoreAssetType{
                            id
                        }
                        ... on IOSStoreAssetType{
                            id
                        }
                        ... on UrlAssetType{
                            id
                        }
                        ... on DomainAssetType{
                            id
                        }
                        ... on SourceCodeAssetType{
                            id
                        }
                        ... on FileAssetType{
                            id
                        }
                        ... on IPAssetType{
                            id
                        }
                        ... on IPv4AssetType{
                            id
                        }
                        ... on IPv6AssetType{
                            id
                        }
                    }
                }
            }
        """

    @property
    def data(self) -> Optional[Dict]:
        """Sets the body of the API request.

        Returns:
            The body of the create asset request.
        """
        variables = {
            'asset': {
                'tags': []
            }
        }
        asset_type_variables = self.__get_asset_variables()
        variables['asset'].update(asset_type_variables)

        if any(isinstance(self._asset, t)
               for t in (android_aab.AndroidAab, android_apk.AndroidApk, file.File, ios_ipa.IOSIpa)):
            map_variables = self._get_map_variables()
            data = {
                'operations': json.dumps({'query': self.query,
                                          'variables': json.dumps(variables)}),
                'map': json.dumps(map_variables)
            }
        else:
            data = {
                'query': self.query,
                'variables': json.dumps(variables)
            }
        return data

    @property
    def files(self) -> Optional[Dict] :
        """Sets the file for multipart upload of the API request.

        Returns:
            The file mapping of the create asset request.
        """
        if any(isinstance(self._asset, t)
               for t in (android_aab.AndroidAab, android_apk.AndroidApk, file.File, ios_ipa.IOSIpa)):
            return {'0': self._asset.content}
        else:
            return None

    def __get_asset_variables(self) -> Dict[str, Dict]:
        """Creates asset variables for the API request, depending on the type of the asset."""
        asset_type_variables = {}
        if isinstance(self._asset, (android_aab.AndroidAab, android_apk.AndroidApk)):
            asset_type_variables = {
                'androidFile': {
                    'file': None
                }
            }
        elif isinstance(self._asset, ios_ipa.IOSIpa):
            asset_type_variables = {
                'iosFile':{
                    'file': None
                }
            }
        elif isinstance(self._asset, file.File):
            asset_type_variables = {
                'file':{
                    'content': None,
                    'path': self._asset.path
                }
            }
        elif isinstance(self._asset, domain_name.DomainName):
            asset_type_variables = {
                'domain':{
                    'domain': self._asset.name
                }
            }
        elif isinstance(self._asset, ip.IP):
            asset_type_variables = {
                'ip':{
                    'host': self._asset.host,
                    'mask': self._asset.mask,
                    'version': self._asset.version
                }
            }
        elif isinstance(self._asset, ipv4.IPv4):
            asset_type_variables = {
                'ipV4':{
                    'host': self._asset.host,
                    'mask': self._asset.mask,
                    'version': self._asset.version
                }
            }
        elif isinstance(self._asset, ipv6.IPv6):
            asset_type_variables = {
                'ipV6':{
                    'host': self._asset.host,
                    'mask': self._asset.mask,
                    'version': self._asset.version
                }
            }
        else:
            raise NotImplementedError(f'Unknown asset type : {type(self._asset)}')

        return asset_type_variables


    def _get_map_variables(self):
        """Creates map variables for the API request, depending on the type of the asset."""
        map_variables = {}
        if isinstance(self._asset, (android_aab.AndroidAab, android_apk.AndroidApk)):
            map_variables = {
                '0': ['variables.asset.androidFile.file']
            }
        elif isinstance(self._asset, ios_ipa.IOSIpa):
            map_variables = {
                '0': ['variables.asset.iosFile.file']
            }
        elif isinstance(self._asset, file.File):
            map_variables = {
                '0': ['variables.asset.file.content']
            }

        return map_variables



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ostorlab/agent/message/proto/v3/report/event/post_scan/timeout/timeout.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\nLostorlab/agent/message/proto/v3/report/event/post_scan/timeout/timeout.proto\x12>ostorlab.agent.message.proto.v3.report.event.post_scan.timeout\"\t\n\x07Message')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ostorlab.agent.message.proto.v3.report.event.post_scan.timeout.timeout_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _MESSAGE._serialized_start=144
  _MESSAGE._serialized_end=153
# @@protoc_insertion_point(module_scope)



"""Report vulnerability Mixin.

Definition of the main method to process an entry from the KB - Knowledge Base,
and emit a vulnerability message.
"""
import dataclasses
import enum
from typing import Optional, List, Any, Union, Dict

from ostorlab.agent.kb import kb
from ostorlab.agent.mixins.protocols import emit
from ostorlab.assets import asset


class MetadataType(enum.Enum):
    """Vulnerability location metadata type."""
    FILE_PATH = enum.auto()
    CODE_LOCATION = enum.auto()
    URL = enum.auto()
    PORT = enum.auto()


@dataclasses.dataclass
class VulnerabilityLocationMetadata:
    """Vulnerability location metadata holding the type
     and the value of 'where' the vulnerability was detected in the asset """
    metadata_type: MetadataType
    value: str


@dataclasses.dataclass
class VulnerabilityLocation:
    """Vulnerability location used to attach a vulnerability to a specific asset"""
    metadata: List[VulnerabilityLocationMetadata]
    asset: asset.Asset

    def to_dict(self) -> Dict[str, Union[Any, List[Dict[str, str]]]]:
        """Convert data class to a dict matching what is expected from protobuf."""
        return {
            self.asset.proto_field: self.asset.__dict__,
            'metadata': [{'value': meta.value, 'type': meta.metadata_type.name} for meta in self.metadata]
        }


class RiskRating(enum.Enum):
    """Risk ratings enumeration."""
    HIGH = enum.auto()
    MEDIUM = enum.auto()
    LOW = enum.auto()
    POTENTIALLY = enum.auto()
    HARDENING = enum.auto()
    SECURE = enum.auto()
    IMPORTANT = enum.auto()
    INFO = enum.auto()


class AgentReportVulnMixin(emit.EmitProtocol):
    """Report Vulnerability class implementing logic of fetching entries from the knowledge base,
    and emitting vulnerability messages."""

    def report_vulnerability(self,
                             entry: kb.Entry,
                             technical_detail: str,
                             risk_rating: RiskRating,
                             dna: Optional[str] = None,
                             vulnerability_location: Optional[VulnerabilityLocation] = None) -> None:
        """Fetch the details of an entry from the knowledge base, and emit a vulnerability message.
        Args:
            entry: knowledge base entry.
            technical_detail: markdown of the scan results.
            risk_rating: the risk rating assigned to the result of the scan.
            dna: unique identifier for duplicate vulnerabilities.
            vulnerability_location: vulnerability location where the vulnerability was detected.
        Returns:
            None
        """
        title = entry.title
        short_description = entry.short_description
        description = entry.description
        recommendation = entry.recommendation
        security_issue = entry.security_issue
        privacy_issue = entry.privacy_issue
        has_public_exploit = entry.has_public_exploit
        targeted_by_malware = entry.targeted_by_malware
        targeted_by_ransomware = entry.targeted_by_ransomware
        targeted_by_nation_state = entry.targeted_by_nation_state
        cvss_v3_vector = entry.cvss_v3_vector

        references = []
        for key, value in entry.references.items():
            reference = {}
            reference['title'] = key
            reference['url'] = value
            references.append(reference)
        data = {
            'title': title,
            'technical_detail': technical_detail,
            'risk_rating': risk_rating.name,
            'short_description': short_description,
            'description': description,
            'recommendation': recommendation,
            'references': references,
            'security_issue': security_issue,
            'privacy_issue': privacy_issue,
            'has_public_exploit': has_public_exploit,
            'targeted_by_malware': targeted_by_malware,
            'targeted_by_ransomware': targeted_by_ransomware,
            'targeted_by_nation_state': targeted_by_nation_state,
            'cvss_v3_vector': cvss_v3_vector,
            'dna': dna
        }
        if vulnerability_location is not None:
            data['vulnerability_location'] = vulnerability_location.to_dict()

        # if dna is not specified, it should not be provided to the portal otherwise it will cause a problem.
        if dna is None:
            data.pop('dna')

        selector = 'v3.report.vulnerability'
        self.emit(selector, data)



"""Retrieve the scan info."""

from typing import Dict, Optional
import json

from ostorlab.apis import request


class ScanInfoAPIRequest(request.APIRequest):
    """Retrieve the scan info."""

    def __init__(self, scan_id: int):
        self._scan_id = scan_id

    @property
    def query(self) -> Optional[str]:
        """Defines the query to get the scan info.

        Returns:
            The query to get the scan info.
        """
        return """
         query scan($scanId: Int) {
            scan(scanId: $scanId) {
                progress
                riskRating
            }
         }
        """

    @property
    def data(self) -> Optional[Dict]:
        """Sets the query to scan info.

        Returns:
              The query to scan info.
        """
        data = {
            'query': self.query,
            'variables': json.dumps({'scanId': self._scan_id})
            }
        return data
